#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <unistd.h>

#include "video_quality.h"

typedef struct {
    FILE *f;
    unsigned width, height;
    unsigned frate;
} vq_stream_t;


static const char *src1 = NULL;
static const char *src2 = NULL;
static unsigned skip1 = 0;
static unsigned skip2 = 0;
static unsigned step1 = 1;
static unsigned step2 = 1;
static unsigned frames = UINT_MAX;
#if 0
static unsigned autoskip = 0;
#endif

#ifndef MPLAYER_STDOUT
# define MPLAYER_STDOUT PREFIX "/libexec/mplayer_stdout"
#endif

static vq_stream_t *
vq_stream_open(const char *video, unsigned swidth, unsigned sheight)
{
    /* XXX */
    char buf[131072];
    vq_stream_t *s;
    unsigned width, height, frate_num, frate_denom;

    if((s = malloc(sizeof(vq_stream_t))) == NULL) {
	perror("malloc");
	exit(1);
    }
    sprintf(buf, MPLAYER_STDOUT " %s", video);
    if(swidth > 0)
	sprintf(buf + strlen(buf), " -vf-add scale=%d:%d", swidth, sheight);
    if((s->f = popen(buf, "r")) == NULL) {
	perror("popen");
	exit(1);
    }
    if(fgets(buf, sizeof(buf), s->f) == NULL) {
	fprintf(stderr, "Error reading video header.\n");
	exit(1);
    }
    if(sscanf(buf, "YUV4MPEG2 W%d H%d F%d:%d", &width, &height,
	&frate_num, &frate_denom) != 4) {
	fprintf(stderr, "Error parsing video header.\n");
	exit(1);
    }
    s->width = width;
    s->height = height;
    s->frate = frate_num / frate_denom;
    return(s);
}

static void
vq_stream_close(vq_stream_t *s)
{
    pclose(s->f);
    free(s);
}

static int
vq_stream_get_frame(vq_stream_t *s, unsigned char *dy, unsigned char *du,
    unsigned char *dv)
{
    unsigned char buf[4096];
    unsigned n;

    if((fgets(buf, sizeof(buf), s->f)) == NULL)
	return(0);
    if(strncmp(buf, "FRAME", 5)) {
	fprintf(stderr, "Error parsing frame header.\n");
	exit(1);
    }
    n = s->width * s->height;
    if(fread(dy, 1, n, s->f) != n) {
	fprintf(stderr, "Error reading frame.\n");
	exit(1);
    }
    n /= 4;
    if(fread(du, 1, n, s->f) != n) {
	fprintf(stderr, "Error reading frame.\n");
	exit(1);
    }
    if(fread(dv, 1, n, s->f) != n) {
	fprintf(stderr, "Error reading frame.\n");
	exit(1);
    }
    return(1);
}

static double
corr_index(unsigned char *d1, unsigned char *d2, unsigned n)
{
    double sx = 0, sy = 0, sxx = 0, syy = 0, sxy = 0;
    unsigned i;

    for(i = 0; i < n; i++) {
	sx += *d1;
	sy += *d2;
	sxx += *d1 * *d1;
	syy += *d2 * *d2;
	sxy += *d1 * *d2;
	d1++;
	d2++;
    }
    return(4 * (n * sxy - sx * sy ) * sx * sy /
	((n * sxx - sx * sx + n * syy - sy * sy) * (sx * sx + sy * sy)));
}

#if 0
static void
handle_autoskip(vq_stream_t *v1, vq_stream_t *v2)
{
    double mse[autoskip][autoskip];
    unsigned char *dy[autoskip + 1], *du[autoskip + 1], *dv[autoskip + 1];
    unsigned ny, nu, nv;
    unsigned i, j;
    double bmse;
    unsigned bi, bj;
    unsigned k1, k2;

    ny = v1->width * v2->height;
    nu = nv = ny / 4;
    if((dy[0] = malloc((ny + nu + nv) * (autoskip + 1))) == NULL) {
	perror("malloc");
	exit(1);
    }
    for(i = 0; i <= autoskip; i++) {
	if(i > 0)
	    dy[i] = dv[i - 1] + nv;
	du[i] = dy[i] + ny;
	dv[i] = du[i] + nu;
    }
    for(i = 0; i < autoskip; i++) {
	if(!vq_stream_get_frame(v1, dy[i], du[i], dv[i])) {
	    fprintf(stderr, "End of 1 during autoskip.\n");
	    exit(1);
	}
    }
    for(j = 0; j < autoskip; j++) {
	if(!vq_stream_get_frame(v2, dy[autoskip], du[autoskip], dv[autoskip])) {
	    fprintf(stderr, "End of 2 during autoskip.\n");
	    exit(1);
	}
	for(i = 0; i < autoskip; i++)
	    mse[i][j] =
		psnr_mse(dy[i], dy[autoskip], ny) +
		psnr_mse(du[i], du[autoskip], nu) +
		psnr_mse(dv[i], dv[autoskip], nv);
    }
    bmse = mse[0][0];
    bi = bj = 0;
    for(i = 0; i < autoskip; i++) {
	for(j = 0; j < autoskip; j++) {
	    if(mse[i][j] < bmse) {
		bmse = mse[i][j];
		bi = i;
		bj = j;
	    }
	}
    }
    k1 = (autoskip - bi + step1 - 1) / step1;
    k2 = (autoskip - bj + step2 - 1) / step2;
    if(k2 > k1)
	k1 = k2;
    k2 = bj + k1 * step2 - autoskip;
    k1 = bi + k1 * step1 - autoskip;
    for(i = 0; i < k1; i++) {
	if(!vq_stream_get_frame(v1, dy[0], du[0], dv[0])) {
	    fprintf(stderr, "End of 1 during autoskip.\n");
	    exit(1);
	}
    }
    for(i = 0; i < k2; i++) {
	if(!vq_stream_get_frame(v2, dy[0], du[0], dv[0])) {
	    fprintf(stderr, "End of 2 during autoskip.\n");
	    exit(1);
	}
    }
    free(dy[0]);
}
#endif

static void
display_frames(unsigned w, unsigned h,
    unsigned char *y1, unsigned char *u1, unsigned char *v1,
    unsigned char *y2, unsigned char *u2, unsigned char *v2)
{
    FILE *f;
    unsigned i;

    f = popen("display pnm:-", "w");
    fprintf(f, "P5\n%d %d\n255\n", w * 2, h + h / 2);
    for(i = 0; i < h; i++) {
	fwrite(y1 + i * w, w, 1, f);
	fwrite(y2 + i * w, w, 1, f);
    }
    w /= 2;
    h /= 2;
    for(i = 0; i < h; i++) {
	fwrite(u1 + i * w, w, 1, f);
	fwrite(v1 + i * w, w, 1, f);
	fwrite(u2 + i * w, w, 1, f);
	fwrite(v2 + i * w, w, 1, f);
    }
    pclose(f);
}

static int
parse_option(const char *o, const char *a, const char **v)
{
    int l;

    l = strlen(o);
    if(memcmp(a, o, l) == 0) {
	if(a[l - 1] == '=') {
	    *v = a + l;
	    return(1);
	}
	if(a[l] == 0)
	    return(1);
    }
    return(0);
}

static void
parse_options(int argc, char **argv)
{
    int i;
    const char *a, *t = NULL;

    for(i = 1; i < argc; i++) {
	a = argv[i];
	if(parse_option("src1=", a, &src1)) {
	    /* nothing */
	} else if(parse_option("src2=", a, &src2)) {
	    /* nothing */
	} else if(parse_option("step=", a, &t)) {
	    step1 = step2 = strtol(t, NULL, 0);
	} else if(parse_option("step1=", a, &t)) {
	    step1 = strtol(t, NULL, 0);
	} else if(parse_option("step2=", a, &t)) {
	    step2 = strtol(t, NULL, 0);
	} else if(parse_option("skip=", a, &t)) {
	    skip1 = skip2 = strtol(t, NULL, 0);
	} else if(parse_option("skip1=", a, &t)) {
	    skip1 = strtol(t, NULL, 0);
	} else if(parse_option("skip2=", a, &t)) {
	    skip2 = strtol(t, NULL, 0);
#if 0
	} else if(parse_option("autoskip=", a, &t)) {
	    autoskip = strtol(t, NULL, 0);
#endif
	} else if(parse_option("frames=", a, &t)) {
	    frames = strtol(t, NULL, 0);
	} else {
	    fprintf(stderr, "Unknown option: %s\n", a);
	    exit(1);
	}
    }
}

int
main(int argc, char **argv)
{
    vq_stream_t *v1, *v2;
    unsigned npv, npy, npu;
    unsigned char *dy1, *dy2, *du1, *du2, *dv1, *dv2;
    unsigned frame = 0;
    unsigned i;
    unsigned tframes;
    vq_psnr_t psnr_y, psnr_u, psnr_v, psnr_yuv;
    vq_correl_t corr_y, corr_u, corr_v, corr_yuv;

    parse_options(argc, argv);
    v1 = vq_stream_open(src1, 0, 0);
    v2 = vq_stream_open(src2, 0, 0);
    if(v2->width != v1->width || v2->height != v1->height) {
	vq_stream_close(v2);
	v2 = vq_stream_open(src2, v1->width, v1->height);
    }
    npy = v1->width * v1->height;
    npu = npv = npy / 4;
    if((dy1 = malloc(2 * (npy + npu + npv))) == NULL) {
	perror("malloc");
	exit(1);
    }
    du1 = dy1 + npy;
    dv1 = du1 + npu;
    dy2 = dv1 + npv;
    du2 = dy2 + npy;
    dv2 = du2 + npu;
    for(i = 0; i < skip1; i++)
	vq_stream_get_frame(v1, dy1, du1, dv1);
    for(i = 0; i < skip2; i++)
	vq_stream_get_frame(v2, dy2, du2, dv2);
#if 0
    if(autoskip > 0)
	handle_autoskip(v1, v2);
#endif

    tframes = 0;
    vq_psnr_init(&psnr_y);
    vq_psnr_init(&psnr_u);
    vq_psnr_init(&psnr_v);
    vq_correl_init(&corr_y);
    vq_correl_init(&corr_u);
    vq_correl_init(&corr_v);

    while(1) {
	if(!vq_stream_get_frame(v1, dy1, du1, dv1))
	    break;
	if(!vq_stream_get_frame(v2, dy2, du2, dv2))
	    break;

	tframes++;
	vq_psnr_feed(&psnr_y, dy1, dy2, npy);
	vq_psnr_feed(&psnr_u, du1, du2, npu);
	vq_psnr_feed(&psnr_v, dv1, dv2, npv);
	vq_correl_feed(&corr_y, dy1, dy2, npy);
	vq_correl_feed(&corr_u, du1, du2, npu);
	vq_correl_feed(&corr_v, dv1, dv2, npv);

	fprintf(stderr, "frame %d\r", ++frame);
	if(--frames == 0)
	    break;
	for(i = 1; i < step1; i++)
	    vq_stream_get_frame(v1, dy1, du1, dv1);
	for(i = 1; i < step2; i++)
	    vq_stream_get_frame(v2, dy2, du2, dv2);
    }
    fprintf(stderr, "\n");

    psnr_yuv = psnr_y;
    vq_psnr_add(&psnr_yuv, 1, &psnr_u, 1);
    vq_psnr_add(&psnr_yuv, 1, &psnr_v, 1);
    corr_yuv = corr_y;
    vq_correl_add(&corr_yuv, 1, &corr_u, 1);
    vq_correl_add(&corr_yuv, 1, &corr_v, 1);

    printf("frames = %d\n", tframes);
    vq_psnr_print(&psnr_y, "_y");
    vq_psnr_print(&psnr_u, "_u");
    vq_psnr_print(&psnr_v, "_v");
    vq_psnr_print(&psnr_yuv, "");
    vq_correl_print(&corr_y, "_y");
    vq_correl_print(&corr_u, "_u");
    vq_correl_print(&corr_v, "_v");
    vq_correl_print(&corr_yuv, "");

    return(0);
}
